/* General includes */
/* Including AVR IO to be able to use PORT, PIN & DDR */
#include <avr/io.h>
/* Include interrupt header, to be able to use sei() to enable interrupts */
#include <avr/interrupt.h>
/* Include delay header to be able to use _delay_ms */
#include <util/delay.h>

#include "xdc.h"




/* ADC channel read currently */
unsigned char xdc_adc_channel = 2;
unsigned char xdc_adc_count = 0;
unsigned long xdc_adc_data[2][5];

unsigned int xdc_voltage_reference = 700; //xxxxxxx !Voltage Limit @ ~1.78V -> 30V maximum voltage
unsigned int xdc_current_reference = 310; //xxxxxxx !Current Limit @ ~0.34V -> 17mA maximum current
unsigned char xdc_voltage_hysteresis = 10;
unsigned char xdc_current_hysteresis = 40; //Allow voltage differences of +/- 2.4mA
unsigned char xdc_timer_top = 255;
unsigned char xdc_timer_safe_operation = 200;
unsigned char xdc_timer_prescaler = 1; //Autostart XDC

volatile unsigned long xdc_middle_voltage = 0;
volatile unsigned long xdc_middle_current = 0;

/* PWM ratio */
unsigned char xdc_p_tast = 1;
unsigned char xdc_enabled = 1;


void xdc_init()
{
	/* Initialize timer 1 for high speed PWM operation */
	PLLCSR |= (1 << PLLE); //Enable high speed (64 MHz PLL)
	_delay_ms(1); //Delay 1ms to be sure that PLL is locked
	PLLCSR |= (1 << PCKE); //Enable PLL clock source for Timer 1

	xdc_enable(xdc_enabled);

	/* Initialize ADC */
	xdc_adcInit(XDC_CHANNEL_VOLTAGE);
	xdc_adc_channel = XDC_CHANNEL_VOLTAGE;
}


void xdc_prefillRegisters()
{
	twiSlaveRegisterWrite(0, xdc_voltage_reference & 0x00FF);
	twiSlaveRegisterWrite(1, xdc_voltage_reference >> 8);
	twiSlaveRegisterWrite(2, xdc_current_reference & 0x00FF);
	twiSlaveRegisterWrite(3, xdc_current_reference >> 8);
	twiSlaveRegisterWrite(4, xdc_voltage_hysteresis);
	twiSlaveRegisterWrite(5, xdc_current_hysteresis);
	twiSlaveRegisterWrite(6, xdc_enabled);
}

void xdc_loadRegisters()
{
	unsigned char old_xdc_enabled;

	old_xdc_enabled 				= xdc_enabled;

	xdc_voltage_reference 	= twiSlaveRegisterRead(0) | (twiSlaveRegisterRead(1) << 8);	
	xdc_current_reference 	= twiSlaveRegisterRead(2) | (twiSlaveRegisterRead(3) << 8);	
	
	xdc_voltage_hysteresis 	= twiSlaveRegisterRead(4);
	xdc_current_hysteresis 	= twiSlaveRegisterRead(5);

	xdc_enabled = twiSlaveRegisterRead(6);

	if(xdc_enabled != old_xdc_enabled)
	{
		xdc_enable(xdc_enabled);
	}

}

void xdc_adjust()
{
	cli();
	xdc_middle_current = xdc_adc_data[XDC_CHANNEL_CURRENT][0] + xdc_adc_data[XDC_CHANNEL_CURRENT][1] + xdc_adc_data[XDC_CHANNEL_CURRENT][2] + xdc_adc_data[XDC_CHANNEL_CURRENT][3] + xdc_adc_data[XDC_CHANNEL_CURRENT][4];
	xdc_middle_voltage = xdc_adc_data[XDC_CHANNEL_VOLTAGE][0] + xdc_adc_data[XDC_CHANNEL_VOLTAGE][1] + xdc_adc_data[XDC_CHANNEL_VOLTAGE][2] + xdc_adc_data[XDC_CHANNEL_VOLTAGE][3] + xdc_adc_data[XDC_CHANNEL_VOLTAGE][4];
	sei();
	
	xdc_middle_current /= 5;
	xdc_middle_voltage /= 5;


	if(xdc_timer_prescaler != 0) //xdc started
	{
		if(xdc_middle_voltage > XDC_ABSOLUTE_MAXIMUM_VOLTAGE || xdc_middle_current > XDC_ABSOLUTE_MAXIMUM_VOLTAGE) //To secure that the chip doesn't get damaged -> restrict absolute maximum output voltage / current
		{
			xdc_ratioAlter(-1);
		}
		/* Limit Voltage */
		else if(xdc_middle_voltage > (xdc_voltage_reference + xdc_voltage_hysteresis))
		{
			xdc_ratioAlter(-1);
		}
		/* Adjust current */
		else if(xdc_middle_current > (xdc_current_reference + xdc_current_hysteresis) || xdc_middle_current < (xdc_current_reference - xdc_current_hysteresis))
		//else
		{
			int x = xdc_current_reference - xdc_middle_current;
			int adjust = 1;
			/*
			if(x > 50 || x < -50)
			{
				adjust = 5;
			}*/
			if(x < 0)
			{
				adjust *= -1;
			}

			xdc_ratioAlter(adjust);
		}
	}
}

void xdc_adcInit(unsigned char channel)
{
	unsigned char admuxv = xdc_adc_channel==XDC_CHANNEL_CURRENT?XDC_CHANNEL_ADC_CURRENT:XDC_CHANNEL_ADC_VOLTAGE;

	if(xdc_adc_channel == XDC_CHANNEL_VOLTAGE)
	{
		DIDR0 = (1 << ADC2D);
		admuxv |= (1 << REFS2) | (1 << REFS1); //2.56V reference
			
	}
	else
	{
		DIDR0 = (1 << ADC3D);
		admuxv |= (1 << REFS1);	//1.1V reference
	}

	ADMUX = admuxv;
	ADCSRA = (1<<ADEN)  | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0)  | (1 << ADIE) | (1 << ADSC)| (1<< ADATE); 	
}

void xdc_adcStop()
{
	ADCSRA = 0;
}

void xdc_adcReinit()
{
	ADCSRA = (1<<ADEN)  | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0)  | (1 << ADIE) | (1 << ADSC)| (1<< ADATE); 		
}


/**
 * Callback if ADC value has been read successfully
 * 
 * @brief AD value read
 * @param value Read ADC value (10bit, 0 - 1024) 
 */
SIGNAL (SIG_ADC)
{
	if(xdc_adc_count != 0) //Ignore first conversion, may be faulty
	{
		xdc_adc_data[xdc_adc_channel][xdc_adc_count-1] = ADC;	
	}
	xdc_adc_count++;

	if(xdc_adc_count > 5)
	{
		xdc_adc_count = 0;
		xdc_adc_channel = !xdc_adc_channel;

		xdc_adcInit(xdc_adc_channel);
	}
}



/**
 * Increase / Decrease PWM ratio
 *
 * @brief Alter PWM ratio
 * @param aDifference Ratio differnence
 */
void xdc_ratioAlter(int aDifference)
{
	if(xdc_timer_prescaler && (xdc_p_tast + aDifference) >= 0 && (xdc_p_tast + aDifference) < xdc_timer_safe_operation)
	{
		xdc_p_tast+=aDifference;
		OCR1A = xdc_p_tast;
	}
}

/**
 * Reset all parameters, so that no harm is done to the outer system
 * If mode is stepdown, disable switching transistor -> no output voltage
 * If mode is stepdown, enable switching transistor -> power transistor disabled -> output voltage equals input voltage
 * @brief Reset PWM & Pin states
 */
void xdc_enable(unsigned char enable)
{
	if(enable)
	{
		/* Set DDR to enable  PWM pin as output */
		DDRB |= (1 << PB1);

		TCCR1 = 0; //Reset timer register
		TCCR1 |= 0b00000001; //Set prescaler to 1
		TCCR1 |= (1 << PWM1A) | (1 << COM1A1); //Set PWM mode
		OCR1C =	xdc_timer_top; //Set reload register
		OCR1A = 0;
		xdc_p_tast = 0;
	}
	else
	{
		TCCR1 = 0; //Reset timer register
		OCR1A = 0; //Disable switching FET, input voltage equals output voltage
		PORTB &= ~(1 << PB1); 
		DDRB &= ~(1 << PB1);
		xdc_p_tast = 0;
	}
}

